/* CGV/solution/parts/cgvp05.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 05 <<<<<<<<<<<<<<<<<<<<<<*/

/************* boundary and exact solution init ****************/

/*------------------------------
 * void Init_Boundary_Vec
 *  OUT vector_chunk b: the boundary vector 
 *
 * TASK: Store all boundary values in the vector memory
 *       (that is already allocated).
 *       In this example, the boundaries are given by:
 *         u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=-1,m and j=-1,n
 *
 * GLOBAL DATA:
 *   IN int start_i : start row of the physical area this process is to calculate
 *   IN int end1_i  : end row + 1 of the physical area this process is to calculate
 *   IN int start_j : start column of the physical area this process is to calculate
 *   IN int end1_j  : end column + 1 of the physical area this process is to calculate
 *
 * USED FUNCTIONS:
 *   APPL:
 *     --
 */
void Init_Boundary_Vec(vector_chunk b)
{
    int    i, j;
    double v, h;

    h = 1.0/(m+1);
    /* boundary vector data is all zero at this point */
    for (i = start_i; i < end1_i; i++) {
        for (j = start_j; j < end1_j; j++) {
            v = 0.0;
            /* next line not neccessary, because boundary value is zero */
            /* if (i==0)   v = v + */ /*u(-1,j):*/ /* h * 0; */    /* = 0 */
            /* next line simplified to add 1.0 to vector, because of the boundary value of 1.0 */
            /* if (i==m-1) v = v + */ /*u(m, j):*/ /* h * (m+1);*/ /* = 1 */
            if (i==m-1) v = 1.0;
            if (j==0)   v = v + /*u(i,-1):*/  h * (i+1);
            if (j==n-1) v = v + /*u(i, n):*/  h * (i+1);
            b[SIDX((i - start_i), (j - start_j))] = v;
        }
    }
}

/*------------------------------
 * void Init_Exact_Solution_Vec
 *   OUT vector_chunk u: the local chunk of the exact solution vector
 *
 * TASK: Store the exact solution for the given boundary in the vector u.
 *       In this example, the inner values are given by:
 *       u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=0..m-1 and j=0..n-1
 *
 * GLOBAL DATA:
 *   IN int start_i : start row of the physical area this process is to calculate
 *   IN int end1_i  : end row + 1 of the physical area this process is to calculate
 *   IN int start_j : start column of the physical area this process is to calculate
 *   IN int end1_j  : end column + 1 of the physical area this process is to calculate
 *   IN int m       : vertical size of the physical problem
 *
 * USED FUNCTIONS:
 *   --
 */
void Init_Exact_Solution_vec(vector_chunk u)
{
    int i, j;
    double h;

    h = 1.0 / (m + 1);
    for (i = start_i; i < end1_i; i++) {
        for (j = start_j; j < end1_j; j++) {
            u[SIDX((i - start_i), (j - start_j))] = h * (i + 1);
        }
    }
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 05 <<<<<<<<<<<<<<<<<<<<<<*/
