/* CGV/solution/parts/cgvp06.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 06 <<<<<<<<<<<<<<<<<<<<<<*/

/******************* print routines ********************************/
void appl_print_vec(char *name, double *v, int mprint, int nprint)
{
    int i,j;
    if (my_rank == 0) {
        printf("%s:       j=",name);
        for (j=0; j<n; j++) {   
            if ((n>nprint ? (nprint-1)*(j+n-2)/(n-1) != (nprint-1)*(j+n-1)/(n-1) : 1)) {
                printf(" %5d",j);
            }
        }
        printf("\n");
        for (i=0; i<m; i++) {
            if ((m>mprint ? (mprint-1)*(i+m-2)/(m-1) != (mprint-1)*(i+m-1)/(m-1) : 1)) {
                printf("%s: i=%3d   ",name,i);
                for (j=0; j<n; j++) {
                    if ((n>nprint ? (nprint-1)*(j+n-2)/(n-1) != (nprint-1)*(j+n-1)/(n-1) : 1)) {
                        printf(" %5.3f",v[n * i +j]);
                    }
                }
                printf("\n");
            }
        }
    }
}

void recv_token(void)
{
#ifndef serial
    int dummy;
    MPI_Status status;
    if (my_rank > 0) {
        MPI_Recv(&dummy, 0, MPI_INT, my_rank-1, MSG_TAG, comm_cart, &status);
    }
#endif
}

void flush_and_send_token(void)
{
#ifndef serial
    MPI_Status status;
    int dummy;
#endif
    fflush(stdout);
#ifndef serial
    MPI_Send(&dummy, 0, MPI_INT, (my_rank+1) % numprocs, MSG_TAG, comm_cart);
    /* Process 0 must wait until the last process (numprocs-1) has finished its output */
    if (my_rank == 0) {
        MPI_Recv(&dummy, 0, MPI_INT, numprocs-1, MSG_TAG, comm_cart, &status);
    }
#endif
}

#ifndef serial
void print_halo()
{
    int i;

    recv_token();

    printf("\nHalo of rank %d:\n", my_rank);
    printf("----------------\n");
    if (halo.north_rank == MPI_PROC_NULL) {
        printf("North: NO halo\n");
    } else {
        printf("North:\n");
        printf("------\n");
        printf("values: %d, neighbor rank: %d\n", halo.north_size, halo.north_rank);
        for (i = 0; i < halo.north_size; i++) {
            printf("halo[%d]: %f\n", i, halo.north[i]);
        }
    }
    if (halo.east_rank == MPI_PROC_NULL) {
        printf("East: NO halo\n");
    } else {
        printf("East:\n");
        printf("------\n");
        printf("values: %d, neighbor rank: %d\n", halo.east_size, halo.east_rank);
        for (i = 0; i < halo.east_size; i++) {
            printf("halo[%d]: %f\n", i, halo.east[i]);
        }
    }
    if (halo.south_rank == MPI_PROC_NULL) {
        printf("South: NO halo\n");
    } else {
        printf("South:\n");
        printf("------\n");
        printf("values: %d, neighbor rank: %d\n", halo.south_size, halo.south_rank);
        for (i = 0; i < halo.south_size; i++) {
            printf("halo[%d]: %f\n", i, halo.south[i]);
        }
    }

    if (halo.west_rank == MPI_PROC_NULL) {
        printf("West: NO halo\n");
    } else {
        printf("West:\n");
        printf("-----\n");
        printf("values: %d, neighbor rank: %d\n", halo.west_size, halo.west_rank);
        for (i = 0; i < halo.west_size; i++) {
            printf("halo[%d]: %f\n", i, halo.west[i]);
        }
    }
    printf("\n");   
    flush_and_send_token();
}

void print_halo_struct()
{
    recv_token();

    printf("\nHalo of rank %d:\n", my_rank);
    printf("----------------\n");
    if (halo.north_rank == MPI_PROC_NULL) {
        printf("North: NO halo\n");
    } else {
        printf("North: values: %d, neighbor rank: %d\n", halo.north_size, halo.north_rank);
    }
    if (halo.east_rank == MPI_PROC_NULL) {
        printf("East: NO halo\n");
    } else {
        printf("East: values: %d, neighbor rank: %d\n", halo.east_size, halo.east_rank);
    }
    if (halo.south_rank == MPI_PROC_NULL) {
        printf("South: NO halo\n");
    } else {
        printf("South: values: %d, neighbor rank: %d\n", halo.south_size, halo.south_rank);
    }
    if (halo.west_rank == MPI_PROC_NULL) {
        printf("West: NO halo\n");
    } else {
        printf("West: values: %d, neighbor rank: %d\n", halo.west_size, halo.west_rank);
    }
    printf("\n");

    flush_and_send_token();
}
#endif
 

void print_vec2d(char *name, vector_chunk v2, int mprint, int nprint)
{
#ifndef serial
    int     i, j, rank;
    double* whole_v2;
    double*     buff;
    int         extent[4], new_offset;
    MPI_Status  status, status2;

    whole_v2 = (double*) safe_malloc(sizeof(double) * n * m);
    buff     = (double*) safe_malloc(sizeof(double) * n * m);

    if (my_rank > 0) {
        /* send the start_i, end1-i, start_j, end1_j */
        extent[0] = start_i;
        extent[1] = end1_i;
        extent[2] = start_j;
        extent[3] = end1_j;
        MPI_Send(extent, 4, MPI_INT, 0, 666, comm_cart);
        
        /* send the vector data to rank 0 */
        MPI_Send(v2, chunksize, MPI_DOUBLE, 0, MSG_TAG, comm_cart);
    } else {
        /* copy the rank 0 chunk into the whole physical vector */
        for (i = start_i; i < end1_i; i++) {
            for (j = start_j; j < end1_j; j++) {
                whole_v2[n * i + j] = v2[SIDX((i - start_i), (j - start_j))];
            }
        }
            
        /* receive the halo data and data extent from all other ranks into the buffer */
        /* and copy the buffer to where it belongs into the whole physical vector */
        for (rank = 1; rank < numprocs; rank++) {
            MPI_Recv(extent, 4, MPI_INT, rank, 666, comm_cart, &status);
            MPI_Recv(buff, ((extent[1] - extent[0]) * (extent[3] - extent[2])), MPI_DOUBLE,
                     rank, MSG_TAG, comm_cart, &status2);
    
            new_offset = extent[3] - extent[2];
            
            for (i = extent[0]; i < extent[1]; i++) {
                for (j = extent[2]; j < extent[3]; j++) {
                    whole_v2[n * i + j] = buff[(i - extent[0]) * new_offset + (j - extent[2])];
                }
            }
        }

        free(buff);
    }
    
    appl_print_vec(name, whole_v2, mprint, nprint);
    free(whole_v2);
#else
    appl_print_vec(name, v2, mprint, nprint); 
#endif
}

void print_vec(char *name, vector_chunk v1)
{
    int i,j;
    recv_token();
    for (i = start_i; i < end1_i; i++) 
      for (j = start_j; j < end1_j; j++) 
        printf("%s[%2d,%2d] = %6.3f\n", name, i,j, v1[SIDX((i - start_i), (j - start_j))]);
    flush_and_send_token();
}

void print_decomp()
{
    recv_token();
    printf("Rank %d: start_i, end1_i, start_j, end1_j: %d, %d, %d, %d\n",
           my_rank, start_i, end1_i, start_j, end1_j);
    flush_and_send_token();
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 06 <<<<<<<<<<<<<<<<<<<<<<*/
