/* CGV/solution/parts/cgvp08.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 08 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ main program *************************/

int main(int argc, char *argv[])
{
    int num_iter;
    double norm_r, norm_err, maxnorm_err;
    vector_chunk u, e;
    double time_start=0, time_end=0;
#ifdef serial
    clock_t time_start_all, time_end_all;
#else
    double time_start_all=1, time_end_all=0;
#endif
    double time_elapsed_all = 1, mflops = 0, global_mflops = 0; 

#ifdef serial
    time_start_all = clock();
    numprocs = 1; my_rank = 0;
#else
    MPI_Init(&argc, &argv);
    time_start_all = MPI_Wtime();
    MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
#endif

    flops = 0;

    m = 4;
    n = 4;
    iter_max = 500;
    epsilon = 1e-6;
    print_level = 1;
    decomp_dims = 1;
    mprt = 11; nprt = 11;
    m_procs=2; n_procs=2;
#ifndef serial
    {
        int dims[2]; dims[0] = 0; dims[1] = 0;
        MPI_Dims_create(numprocs, 2, dims);
        m_procs = dims[0]; n_procs = dims[1];
    }
#endif

    /* Parse command line options */
    read_options( argc, argv );

    /* Init the domain decomposition */
    Init_Decomp();
    if (print_level >= 1) {
        print_decomp();
    }

#ifndef serial
    /* !!! determine communication partners for vector x !!! */
    Init_Halo_Struct();
    if (print_level>=4) {
        print_halo_struct();
    }
#endif

    /* reserve chunk of vector b and !!! assign border data */
    b = alloc_vec();
    Init_Boundary_Vec(b);
    if (print_level>=4) {
        print_vec("b",b);
    }

    /* x is initialized with zeros */
    x = alloc_vec();

#ifndef serial
    time_start = MPI_Wtime();
#endif

    CG_solve(&num_iter, &norm_r);

#ifndef serial
    time_end = MPI_Wtime();
#endif

    u = alloc_vec();
    e = alloc_vec();
    Init_Exact_Solution_vec(u);
    /* p = u - x */
    add_vec_vec(u, -1, x, e);
    norm_err = sqrt(sqr_norm_vec(e));
    maxnorm_err = max_norm_vec(e);

    if (print_level>=3) {
        print_vec2d("x",x,m,n);
    }

#ifndef serial
    time_end_all = MPI_Wtime();
    time_elapsed_all = time_end_all - time_start_all;
    mflops = flops / (1000000 * time_elapsed_all);
    MPI_Reduce(&mflops, &global_mflops, 1, MPI_DOUBLE, MPI_SUM, 0, comm_cart);
#else
    time_end_all = clock();
    time_elapsed_all = ((double) (time_end_all - time_start_all)) / CLOCKS_PER_SEC;
    mflops = flops / (1000000 * time_elapsed_all);
    global_mflops = mflops;
#endif

    if (my_rank==0) {
        printf("main: %3d iterations, sqrt(eps) = %8.2e,  normalized: %8.2e\n",num_iter,
                sqrt(epsilon), sqrt(epsilon/(n*m)));
        printf("main:                 ||r||     = %8.2e,  normalized: %8.2e\n",
                norm_r, norm_r/sqrt(n*m));
        printf("main:                 ||error|| = %8.2e,  normalized: %8.2e\n",
                norm_err, norm_err/sqrt(n*m));
        printf("main:                 max(error)= %8.2e = normalized: %8.2e\n",
                maxnorm_err, maxnorm_err);
        printf("main:                 n*m = %d*%d = %d\n",n, m, n*m);
        printf("main:                 %8.2e sec\n",time_elapsed_all);

        printf("Calculated MFLOPS (avg):  %10.6f\n", global_mflops / numprocs);
        printf("Calculated MFLOPS (all):  %10.6f\n", global_mflops);
    }

    if (print_level>=1) {
        print_vec2d("x",x,mprt,nprt);
    }

#ifndef serial
    MPI_Finalize();
#endif
    return 0;
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 08 <<<<<<<<<<<<<<<<<<<<<<*/
